from __future__ import print_function, absolute_import


class Route:
    """
    Apipie Route
    """

    def __init__(self, path, method, description=""):
        self.path = path
        self.method = method.lower()
        self.description = description

    @property
    def params_in_path(self):
        return [part[1:] for part in self.path.split('/') if part.startswith(':')]

    def path_with_params(self, params=None):
        if params is None:
            return self.path
        else:
            result = self.path
            for param in self.params_in_path:
                if param in params:
                    result = result.replace(':{}'.format(param), str(params[param]))
                else:
                    raise KeyError
            return result



from __future__ import print_function, absolute_import

from apypie.route import Route
from apypie.example import Example
from apypie.param import Param
from apypie.exceptions import MissingArgumentsError, InvalidArgumentTypesError


class Action:
    """
    Apipie Action
    """

    def __init__(self, name, resource, api):
        self.name = name
        self.resource = resource
        self.api = api

    @property
    def apidoc(self):
        resource_methods = self.api.apidoc['docs']['resources'][self.resource]['methods']
        return [method for method in resource_methods if method['name'] == self.name][0]

    @property
    def routes(self):
        return [Route(route['api_url'], route['http_method'], route['short_description']) for route in self.apidoc['apis']]

    @property
    def params(self):
        return [Param(**param) for param in self.apidoc['params']]

    @property
    def examples(self):
        return [Example.parse(example) for example in self.apidoc['examples']]

    def call(self, params={}, headers={}, options={}):
        return self.api.call(self.resource, self.name, params, headers, options)

    def find_route(self, params=None):
        params = self.filter_empty_params(params)
        sorted_routes = sorted(self.routes, key=lambda route: [-1 * len(route.params_in_path), route.path])
        for route in sorted_routes:
            if sorted(route.params_in_path) == sorted(params.keys()):
                return route
        return sorted_routes[-1]

    def validate(self, values=None):
        self._validate(self.params, values)

    def _validate(self, params, values, path=None):
        given_params = set(self.filter_empty_params(values).keys())
        required_params = set([param.name for param in params if param.required])
        missing_params = required_params - given_params
        if missing_params:
            raise MissingArgumentsError(missing_params)

        for param, value in values.items():
            param_descriptions = [p for p in params if p.name == param]
            if param_descriptions:
                param_description = param_descriptions[0]
                if param_description.params and value is not None:
                    if param_description.expected_type == 'array':
                        for item in value:
                            # None = add_to_path(path, param_description.name, i)
                            self._validate(param_description.params, item, None)
                    if param_description.expected_type == 'hash':
                        # None = add_to_path(path, param_description.name)
                        self._validate(param_description.params, value, None)

    def filter_empty_params(self, params=None):
        if params is not None:
            if isinstance(params, dict):
                return dict((k, v) for k, v in params.items() if v is not None)
            else:
                raise InvalidArgumentTypesError
        else:
            return {}



from __future__ import print_function, absolute_import

import json
import os
import errno
try:
    from urlparse import urljoin
except ImportError:
    from urllib.parse import urljoin
import requests

from apypie.resource import Resource
from apypie.exceptions import DocLoadingError


class Api:
    """
    Apipie API bindings

    :param uri: base URL of the server
    :param username: username to access the API
    :param password: username to access the API
    :param api_version: version of the API. Defaults to `1`
    :param language: prefered locale for the API description
    :param apidoc_cache_base_dir: base directory for building apidoc_cache_dir. Defaults to `~/.cache/apipie_bindings`.
    :param apidoc_cache_dir: where to cache the JSON description of the API. Defaults to `apidoc_cache_base_dir/<URI>`.
    :param apidoc_cache_name: name of the cache file. If there is cache in the `apidoc_cache_dir`, it is used. Defaults to `default`.
    :param verify_ssl: should the SSL certificate be verified. Defaults to `True`.

    Usage::

      >>> import apypie
      >>> api = apypie.Api(uri='https://api.example.com', username='admin', password='changeme')
    """

    def __init__(self, **kwargs):
        self.uri = kwargs.get('uri')
        self.api_version = kwargs.get('api_version', 1)
        self.language = kwargs.get('language')
        apidoc_cache_base_dir = kwargs.get('apidoc_cache_base_dir', os.path.join(os.path.expanduser('~/.cache'), 'apypie'))
        self.apidoc_cache_dir = kwargs.get('apidoc_cache_dir', os.path.join(apidoc_cache_base_dir, self.uri.replace(':', '_').replace('/', '_'), 'v{}'.format(self.api_version)))
        self.apidoc_cache_name = kwargs.get('apidoc_cache_name', 'default')

        self._session = requests.Session()
        self._session.verify = kwargs.get('verify_ssl', True)

        self._session.headers = {
            'Accept': 'application/json;version={}'.format(self.api_version)
        }
        if self.language:
            self._session.headers['Accept-Language'] = self.language

        if kwargs.get('username') and kwargs.get('password'):
            self._session.auth = (kwargs['username'], kwargs['password'])

        self.apidoc = self._load_apidoc()

    @property
    def resources(self):
        """List of available resources.

        Usage::

            >>> api.resources
            ['comments', 'users']
        """
        return sorted(self.apidoc['docs']['resources'].keys())

    def resource(self, name):
        """
        Get a resource.

        :param name: the name of the resource to load
        :return: :class:`Resource <Resource>` object
        :rtype: apypie.Resource

        Usage::

            >>> api.resource('users')
        """
        if name in self.resources:
            return Resource(self, name)
        else:
            raise KeyError

    def _load_apidoc(self):
        apifile = os.path.join(self.apidoc_cache_dir, '{}.json'.format(self.apidoc_cache_name))
        try:
            with open(apifile, 'r') as f:
                api_doc = json.load(f)
        except IOError:
            api_doc = self._retrieve_apidoc(apifile)
        return api_doc

    def _retrieve_apidoc(self, apifile):
        try:
            os.makedirs(self.apidoc_cache_dir)
        except OSError as err:
            if err.errno != errno.EEXIST or not os.path.isdir(self.apidoc_cache_dir):
                raise
        try:
            response = self._retrieve_apidoc_call('/apidoc/v{}.json'.format(self.api_version))
        except Exception as e:
            raise DocLoadingError("""Could not load data from {0}: {1}
              - is your server down?
              - was rake apipie:cache run when using apipie cache? (typical production settings)""".format(self.uri, e))
        with open(apifile, 'w') as f:
            f.write(json.dumps(response))
        return response

    def _retrieve_apidoc_call(self, path):
        return self.http_call('get', path)

    def call(self, resource_name, action_name, params={}, headers={}, options={}):
        """
        Call an action in the API.

        It finds most fitting route based on given parameters
        with other attributes necessary to do an API call.

        :param resource_name: name of the resource
        :param action_name: action_name name of the action
        :param params: Dict of parameters to be send in the request
        :param headers: Dict of headers to be send in the request
        :param options: Dict of options to influence the how the call is processed
           * `skip_validation` (Bool) *false* - skip validation of parameters
        :return: :class:`dict` object
        :rtype: dict

        Usage::

            >>> api.call('users', 'show', {'id': 1})
        """
        resource = Resource(self, resource_name)
        action = resource.action(action_name)
        if not options.get('skip_validation', False):
            action.validate(params)

        return self._call_action(action, params, headers, options)

    def _call_action(self, action, params={}, headers={}, options={}):
        route = action.find_route(params)
        get_params = dict((key, value) for key, value in params.items() if key not in route.params_in_path)
        return self.http_call(
            route.method,
            route.path_with_params(params),
            get_params,
            headers, options)

    def http_call(self, http_method, path, params=None, headers=None, options=None):
        full_path = urljoin(self.uri, path)
        kwargs = {}

        if headers:
            kwargs['headers'] = headers

        if params:
            if http_method == 'get':
                kwargs['params'] = params
            else:
                kwargs['json'] = params

        request = self._session.request(http_method, full_path, **kwargs)
        request.raise_for_status()
        return request.json()



from __future__ import print_function, absolute_import

from apypie.action import Action


class Resource:
    """
    Apipie Resource
    """

    def __init__(self, api, name):
        self.api = api
        self.name = name

    @property
    def actions(self):
        return sorted([method['name'] for method in self.api.apidoc['docs']['resources'][self.name]['methods']])

    def action(self, name):
        if self.has_action(name):
            return Action(name, self.name, self.api)
        else:
            raise KeyError

    def has_action(self, name):
        return name in self.actions

    def call(self, action, params={}, headers={}, options={}):
        return self.api.call(self.name, action, params, headers, options)
